from django.contrib.sites.models import Site as DjangoSite, SiteManager
from django.core.exceptions import ImproperlyConfigured
from django.db import models
from django.http.request import split_domain_port


class Site(DjangoSite):
    settings = models.JSONField(blank=True, default=dict)
    port = models.PositiveIntegerField(null=True)
    protocol = models.CharField(default='http', max_length=5)

    @property
    def url(self):
        return f'{self.protocol}://{self.domain}'


old_get_current = SiteManager.get_current
DjangoSite.objects.model = Site


def get_current(self, request=None):
    try:
        return old_get_current(self, request)
    except (ImproperlyConfigured, Site.DoesNotExist):
        if not request:
            return Site(domain='localhost', name='localhost')
        host = request.get_host()
        domain, port = split_domain_port(host)
        protocol = request.META['wsgi.url_scheme']
        Site.objects.create(
            name=domain.capitalize(),
            domain=host,
            port=port or 443 if protocol == 'https' else 80,
            protocol=protocol,
        )
    return old_get_current(self, request)


SiteManager.get_current = get_current



"""
Crudlfa+ PasswordView, Become and BecomeUser views.

Crudlfa+ takes views further than Django and are expected to:

- generate their URL definitions and reversions,
- check if a user has permission for an object,
- declare the names of the navigation menus they belong to.
"""

import logging

from django import http
from django.contrib import auth, messages
from django.contrib.auth.forms import PasswordChangeForm, SetPasswordForm
from django.utils.translation import ugettext_lazy as _

from crudlfap import shortcuts as crudlfap

logger = logging.getLogger()


class PasswordView(crudlfap.UpdateView):
    slug = 'password'
    icon = 'vpn_key'
    color = 'purple'
    controller = 'modal'
    action = 'click->modal#open'

    def get_title_submit(self):
        return _('update').capitalize()

    def get_form_class(self):
        if self.object == self.request.user:
            cls = PasswordChangeForm
        else:
            cls = SetPasswordForm
        # This fixes the form messages feature from UpdateView
        return type(cls.__name__, (cls,), dict(instance=self.object))

    def get_form_kwargs(self):
        kwargs = super().get_form_kwargs()
        kwargs['user'] = self.object
        return kwargs


class BecomeUser(crudlfap.ObjectView):
    urlname = 'su'
    menus = ['object']
    icon = 'accessibility_new'
    color = 'purple'
    link_attributes = {'data-noprefetch': 'true'}

    def become(self):
        """
        Implement this method to override the default become logic.
        """
        auth.login(self.request, self.object, backend=self.backend)

    def get_title_menu(self):
        return _('become').capitalize()

    def get_object(self, queryset=None):
        user = super().get_object()

        if user:
            user.backend = self.backend
        else:
            messages.error(
                self.request,
                'Could not find user {}'.format(self.kwargs['username'])
            )

        return user

    def get(self, request, *a, **k):
        logger.info('BecomeUser by {}'.format(self.request.user))
        request_user = request.user
        self.request = request
        result = self.become()
        if 'become_user' not in request.session:
            request.session['become_user_realname'] = str(request_user)
            request.session['become_user'] = str(request_user.pk)
        messages.info(
            request,
            _('Switched to user %s') % request.user
        )
        return result if result else http.HttpResponseRedirect(
            '/' + self.router.registry.urlpath)

    def get_backend(self):
        return 'django.contrib.auth.backends.ModelBackend'


class Become(crudlfap.View):
    urlname = 'su'

    def has_perm(self):
        return 'become_user' in self.request.session

    def get_backend(self):
        return 'django.contrib.auth.backends.ModelBackend'

    def get_object(self):
        user = self.model.objects.get(pk=self.request.session['become_user'])
        user.backend = self.backend
        return user

    def get(self, request, *a, **k):
        logger.info('Become by {}'.format(self.request.user))
        if 'become_user' not in request.session:
            logger.debug(
                'No become_user in session {}'.format(self.request.user))
            return http.HttpResponseNotFound()

        user = self.get_object()
        auth.login(request, user)
        messages.info(
            request,
            _('Switched back to your user %s') % user
        )
        return http.HttpResponseRedirect('/' + self.registry.urlpath)



"""Sets the default appconfig to :py:cls:`~crudlfap.apps.DefaultConfig`."""
from django.utils.module_loading import autodiscover_modules


def autodiscover():
    autodiscover_modules('crudlfap')


default_app_config = 'crudlfap.apps.DefaultConfig'



import json

from django import http
from django.contrib import messages


class FormMixin:
    """Mixin for views which have a Form."""

    success_url_next = True
    initial = {}
    default_template_name = 'crudlfap/form.html'

    def get_context(self, **context):
        context['form'] = self.form
        return super().get_context(**context)

    def form_valid(self):
        """If the form is valid, redirect to the supplied URL."""
        if self.request.content_type == 'application/json':
            return self.form_valid_json()

        self.message_success()
        return http.HttpResponseRedirect(self.success_url)

    def form_valid_json(self):
        return http.JsonResponse({
            'data': self.form.cleaned_data,
            'status': 'accepted',
        }, status=201)

    def form_invalid_json(self):
        data = dict(
            status='invalid data',
            non_field_errors=self.form.non_field_errors(),
            field_errors=dict(),
        )
        for name, errors in self.form.errors.items():
            data['field_errors'][name] = errors
        return http.JsonResponse(data, status=405)

    def form_invalid(self):
        """If the form is invalid, render the invalid form."""
        if self.request.content_type == 'application/json':
            return self.form_invalid_json()

        self.message_error()
        response = self.render_to_response()
        response.status_code = 400
        return response

    def get_title_submit(self):
        """
        Title of the submit button.

        Defaults to :py:attr:`~crudlfap.mixins.menu.MenuMixin.title_menu`
        """
        return self.title_menu

    def post(self, request, *args, **kwargs):
        """
        Handle POST requests: instantiate a form instance with the passed
        POST variables and then check if it's valid.
        """
        if self.form.is_valid():
            return self.form_valid()
        else:
            return self.form_invalid()

    def get_initial(self):
        """Return the initial data to use for forms on this view."""
        return self.initial.copy()

    def get_prefix(self):
        """Return the prefix to use for forms."""
        return None

    def get_form_class(self):
        if self.router:
            return getattr(self.router, 'form_class')

    def get_form(self):
        """Return an instance of the form to be used in this view."""
        self.form = self.form_class(**self.get_form_kwargs())
        return self.form

    def get_form_kwargs(self):
        """Return the keyword arguments for instantiating the form."""
        self.form_kwargs = {
            'initial': self.get_initial(),
            'prefix': self.get_prefix(),
        }

        if self.request.method in ('POST', 'PUT'):
            if self.request.content_type == 'application/json':
                data = json.loads(self.request.body)
            else:
                data = self.request.POST
            self.form_kwargs.update({
                'data': data,
                'files': self.request.FILES,
            })
        return self.form_kwargs

    def get_next_url(self):
        if '_next' in self.request.POST:
            self.next_url = self.request.POST.get('_next')
        if '_next' in self.request.GET:
            self.next_url = self.request.GET.get('_next')

    def get_success_url(self):
        if self.next_url:
            return self.next_url

        if (hasattr(self, 'object')
                and hasattr(self.object, 'get_absolute_url')):
            return self.object.get_absolute_url()

        if self.router['list']:
            return self.router['list'].url

        return super().get_success_url()

    def message_html(self, message):
        return message

    def message_success(self):
        messages.success(
            self.request,
            self.message_html(self.form_valid_message)
        )

    def message_error(self):
        messages.error(
            self.request,
            self.message_html(self.form_invalid_message)
        )

    def get_swagger_summary(self):
        return self.title

    def get_swagger_post(self):
        result = {
            'consumes': ['application/json'],
            # 'description': self.title,
            # 'operationId': 'addPet',
            'parameters': [],
            'produces': ['application/json', 'application/xml'],
            'responses': {
                '405': {
                    'description': 'Invalid input'
                },
            },
            'summary': self.swagger_summary,
            'tags': self.swagger_tags,
        }
        for name, field in self.form_class.base_fields.items():
            result['parameters'].append({
                'description': '. '.join([
                    str(field.label),
                    str(field.help_text)
                ]),
                'name': name,
                'required': field.required,
                # 'schema': {'$ref': '#/definitions/Pet'}
            })
        return result



from django.apps import apps
from django.conf import settings
from django.contrib.auth.models import Group
from django.urls import reverse

from crudlfap import shortcuts as crudlfap
from crudlfap import html

from . import views

User = apps.get_model(getattr(settings, 'AUTH_USER_MODEL', 'auth.User'))


def login_logout(request, menu):
    if request.user.is_authenticated:
        menu.append(html.MDCListItem(
            'Logout',
            icon='logout',
            href=reverse('logout'),
            style='text-decoration: none',
            up_target=html.UNPOLY_TARGET_ALL,
            tag='a',
        ))
    else:
        menu.insert(1, html.MDCListItem(
            'Login',
            icon='login',
            href=reverse('login'),
            style='text-decoration: none',
            up_target=html.UNPOLY_TARGET_ALL,
            tag='a',
        ))
    return menu


html.mdcDrawer.menu_hooks.append(login_logout)


crudlfap.Router(
    User,
    views=[
        crudlfap.DeleteObjectsView,
        crudlfap.DeleteView,
        crudlfap.UpdateView.clone(
            fields=[
                'username',
                'email',
                'first_name',
                'last_name',
                'groups',
                'is_superuser',
                'is_staff',
                'is_active',
            ]
        ),
        crudlfap.CreateView.clone(
            fields=[
                'username',
                'email',
                'groups',
                'is_staff',
                'is_superuser'
            ],
        ),
        views.PasswordView,
        views.BecomeUser,
        crudlfap.DetailView.clone(exclude=['password']),
        crudlfap.ListView.clone(
            search_fields=[
                'username',
                'email',
                'first_name',
                'last_name'
            ],
            table_fields=[
                'username',
                'email',
                'first_name',
                'last_name',
                'is_staff',
                'is_superuser'
            ],
            filter_fields=[
                'groups',
                'is_superuser',
                'is_staff'
            ],
        ),
    ],
    urlfield='username',
    icon='person',
).register()


class GroupUpdateView(crudlfap.UpdateView):
    def get_form_class(self):
        cls = super().get_form_class()
        cls.base_fields['permissions'].queryset = (
            cls.base_fields['permissions'].queryset.select_related(
                'content_type'))
        return cls


crudlfap.Router(
    Group,
    fields=['name', 'permissions'],
    icon='group',
    views=[
        crudlfap.DeleteObjectsView,
        crudlfap.DeleteView,
        GroupUpdateView,
        crudlfap.CreateView,
        crudlfap.DetailView,
        crudlfap.ListView,
    ],
).register()

crudlfap.site.views.append(views.Become.clone(model=User))



# Generated by Django 3.2 on 2021-04-10 15:30

from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ('crudlfap_sites', '0001_initial'),
    ]

    operations = [
        migrations.AlterField(
            model_name='site',
            name='settings',
            field=models.JSONField(blank=True, default=dict),
        ),
    ]



from django.core.exceptions import ImproperlyConfigured
from django.utils.translation import ugettext as _


class ModelMixin(object):
    """Mixin for views using a Model class but no instance."""
    model = None

    menus = ['model']
    menus_display = ['model']
    pluralize = False

    def get_exclude(self):
        return []

    def get_fields(self):
        return [
            f for f in self.router.get_fields(self)
            if self.model._meta.get_field(f).editable
            and f not in self.exclude
        ]

    def get_model_verbose_name(self):
        if self.pluralize:
            return self.model._meta.verbose_name_plural
        else:
            return self.model._meta.verbose_name

    def get_title(self):
        """Compose a title of Model Name: View label."""
        return '{}: {}'.format(
            self.model_verbose_name.capitalize(),
            _(self.view_label),
        ).capitalize()

    def get_queryset(self):
        """Return router.get_queryset() by default, otherwise super()."""
        if self.router:
            return self.router.get_queryset(self)

        if self.model:
            return self.model._default_manager.all()

        raise ImproperlyConfigured(
            "%(cls)s is missing a QuerySet. Define "
            "%(cls)s.model, %(cls)s.queryset, or override "
            "%(cls)s.get_queryset()." % {
                'cls': self.__class__.__name__
            }
        )

    def get_swagger_model_name(self):
        if self.model:
            return self.model.__name__

    def get_swagger_tags(self):
        if self.model:
            return [self.model.__name__]
        return []



import pytest
from django.db import models
from django.urls import resolve, reverse
from django.views import generic

from crudlfap import shortcuts as crudlfap
from crudlfap_example.artist.models import Artist

from .route import Route
from .router import Router


class TestRouter(Router):
    __test__ = False
    fields = '__all__'


def test_views():
    router = TestRouter()
    view = crudlfap.TemplateView.clone(urlname='home')
    router.views = [view]
    assert router.views['home'] == view
    assert router.views[0].urlname == 'home'

    router.views['home'] = router.views['home'].clone(template_name='lol')
    assert router.views['home'].template_name == 'lol'

    del router.views['home']
    assert len(router.views) == 0


def test_urlfield():
    assert Router().urlfield is None


def test_urlfield_with_model():
    class TestModel(models.Model):
        pass
    assert Router(model=TestModel).urlfield == 'pk'


def test_urlfield_with_slug():
    class TestModel(models.Model):
        slug = models.CharField(max_length=100)
    assert Router(model=TestModel).urlfield == 'slug'


def test_namespace_none():
    assert Router().namespace is None


def test_namespace():
    assert Router(namespace='lol').namespace == 'lol'


def test_namespace_with_model():
    assert Router(model=Artist).namespace == 'artist'


def test_path():
    assert Router(urlpath='foo').urlpath == 'foo'


def test_path_with_model():
    assert Router(model=Artist).urlpath == 'artist'


def test_app_name():
    assert Router(app_name='lol').app_name == 'lol'


def test_app_name_with_model():
    assert Router(model=Artist).app_name == 'artist'


def test_registry_default():
    assert Router().registry == crudlfap.site


def test_registry():
    site = dict()
    assert Router(registry=site).registry == site


class DetailView(Route, generic.DetailView):
    menus = ['object']

    # Not setting this would require
    # request.user.has_perm('artist.detail_artist', obj) to pass
    allowed = True

    # This is done by crudlfap generic ObjectView, but here tests django
    # generic views
    def get_urlargs(self):
        # This may be executed with just the class context (self.object
        # resolving to type(self).object, as from
        # Route.clone(object=something).url
        return [self.object.name]


@pytest.fixture
def router():
    return Router(model=Artist, views=[DetailView])


def test_getitem(router):
    assert issubclass(router['detail'], DetailView)


def test_urlpatterns(router):
    assert len(router.urlpatterns) == 1
    assert router.urlpatterns[0].name == 'detail'


def test_urlpattern(router):
    assert reverse('detail', router) == '/detail'
    assert resolve('/detail', router).func.view_class == router.views[0]


@pytest.mark.django_db
def test_get_menu(router, srf):
    a = Artist(name='a')
    from crudlfap_auth.crudlfap import User
    srf.user = User.objects.create(is_superuser=True)
    req = srf.get('/')
    result = router.get_menu('object', req, object=a)
    assert len(result) == 1
    assert isinstance(result[0], DetailView)
    assert result[0].urlargs == ['a']
    assert type(result[0]).urlargs == ['a']
    assert str(result[0].url) == '/artist/a'

    b = type(result[0]).clone(object=a)
    assert str(b.url) == '/artist/a'



import os

from django import http
from crudlfap.views.generic import View


schema = {
    'basePath': '/v2',
    'definitions': {},
    'info': {},
    'paths': {},
    'schemes': ['https', 'http'],
    'securityDefinitions': {},
    'swagger': '2.0',
    'tags': []
}


class SchemaView(View):
    authenticate = False

    def get(self, request, *args, **kwargs):
        schema['definitions'] = dict()
        schema['paths'] = dict()
        schema['host'] = request.get_host()
        for model, router in self.registry.items():
            for view in router.views:
                view = view.abstract(request=request)
                if not view.has_perm():
                    continue
                if getattr(view, 'router', None):
                    self.add_router_schema(request, view.router, schema)
                try:
                    path_definition = view.swagger_path_definition
                except AttributeError:
                    pass
                else:
                    if path_definition:
                        self.add_path_definition(path_definition, view, schema)
        return http.JsonResponse(schema)

    def add_path_definition(self, path_definition, view, schema):
        url = view.urlpath
        if router := getattr(view, 'router', None):
            if url:
                url = os.path.join(router.urlpath, url)
            else:
                url = router.urlpath
        if self.registry.urlpath:
            url = self.registry.urlpath + url
        url = '/' + url.replace('<', '{').replace('>', '}')
        schema['paths'][url] = path_definition

    def add_router_schema(self, request, router, schema):
        try:
            model_name = router.get_swagger_model_name(request)
        except AttributeError:
            pass
        else:
            if model_name and model_name not in schema['definitions']:
                model_definition = router.get_swagger_model_definition(request)
                if model_definition:
                    schema['definitions'][model_name] = model_definition



from django.urls import reverse

from crudlfap import html


def registration(request, menu):
    if not request.user.is_authenticated:
        menu.insert(1, html.A(
            html.MDCListItem('Signup', icon='badge'),
            href=reverse('django_registration_register'),
            style='text-decoration: none',
        ))
    return menu


html.mdcDrawer.menu_hooks.append(registration)



from crudlfap.models import URL, Controller
from crudlfap.router import Router
from crudlfap.views import generic


class ControllerRouter(Router):
    model = Controller

    views = [
        generic.DetailView,
        generic.ListView.clone(
            search_fields=(
                'app',
                'model',
            ),
            table_fields=(
                'app',
                'model',
            ),
        ),
    ]


# useless ?
# ControllerRouter().register()


class URLRouter(Router):
    model = URL
    icon = 'link'

    views = [
        generic.DetailView,
        generic.ListView.clone(
            search_fields=(
                'name',
                'controller__app',
                'controller__model',
            ),
            table_fields=(
                'controller',
                'id',
                'fullurlpath',
            ),
        ),
    ]


URLRouter().register()



from django.conf import settings
from django.urls import include, path, re_path

from crudlfap import shortcuts as crudlfap

urlpatterns = [
    crudlfap.site.urlpattern,
    path('auth/', include('django.contrib.auth.urls')),
    path('bundles/', include('ryzom_django.bundle')),
]

# CRUDLFA+ extras
if 'crudlfap_registration' in settings.INSTALLED_APPS:
    urlpatterns.append(
        path(
            'registration/',
            include('django_registration.backends.activation.urls')
        ),
    )

if 'debug_toolbar' in settings.INSTALLED_APPS and settings.DEBUG:
    import debug_toolbar
    urlpatterns += [
        re_path(r'^__debug__/', include(debug_toolbar.urls)),
    ]

# PUBLIC DEMO mode
from django.contrib.auth.views import LoginView  # noqa
from django.contrib.auth.models import User  # noqa


def login_view(request, *args, **kwargs):
    if request.method == 'POST':
        for name in ('admin', 'staff', 'user'):
            user = User.objects.filter(username=name).first()
            if not user:
                user = User(username=name, email='user@example.com')
            if name == 'staff':
                user.is_staff = True
            if name == 'admin':
                user.is_superuser = True
            user.set_password(name)
            user.save()
    return LoginView.as_view()(request, *args, **kwargs)
urlpatterns.insert(0, path('auth/login/', login_view, name='login'))  # noqa


from crudlfap_auth.html import *  # noqa
@template('registration/login.html', App)  # noqa
class DemoLogin(LoginFormViewComponent):
    def to_html(self, *content, **context):
        return super().to_html(
            H3('Demo mode enabled'),
            P('Login with either username/password of:'),
            Ul(
                Li('user/user: for user'),
                Li('staff/staff: for staff'),
                Li('admin/admin: for superuser (technical stuff should appear)'),  # noqa
            ),
            *content,
            **context,
        )



from django.conf import settings
from django.contrib.sites.models import Site
from django.urls import reverse
from django.utils.translation import ugettext as _
from ryzom_django_mdc.html import *  # noqa


UNPOLY_TARGET_ALL = '#main, .mdc-top-app-bar__title, #drawer .mdc-list'


class A(A):
    attrs = dict(
        up_target=UNPOLY_TARGET_ALL,
        # up_transition='cross-fade',
    )


class Form(Form):
    attrs = dict(
        up_target=A.attrs['up-target'],
        method='post',
    )


class Container(Div):
    sass = '''
    .Container
        display: flex
        flex-direction: column
        align-items: center
        max-width: 960px
        width: 96vw
        margin: auto

    .up-modal .Container
        width: 86vw

    @media screen and (max-width: 500px)
        .up-modal .up-modal-content
            padding: 10px
    '''


class FormContainer(Container):
    sass = '''
    .FormContainer
        max-width: 580px
        .mdc-text-field, .mdc-form-field, .mdc-select, form
            width: 100%
            max-width: 90vw
    '''


class PageMenu(Div):
    attrs = dict(cls='mdc-elevation--z2', style='margin-bottom: 10px')

    def __init__(self, *args, _next=None, **kwargs):
        super().__init__(*args, **kwargs)
        self._next = _next

    def to_html(self, *content, **context):
        if 'page-menu' not in context:
            return super().to_html(*content, **context)

        content = list(content)
        menu = context['page-menu']

        for v in menu:
            if v.urlname == context['view'].urlname:
                continue

            href = v.url
            if self._next:
                if '?' not in href:
                    href += '?'
                href += '&_next=' + self._next
            button = A(
                MDCTextButton(
                    v.label.capitalize(),
                    icon=getattr(v, 'icon', None),
                    tag='span',
                    style={
                        'margin': '10px',
                        'color': getattr(v, 'color', 'inherit'),
                    },
                ),
                href=href,
                style='text-decoration: none',
            )
            if getattr(v, 'controller', None) == 'modal':
                button.attrs.up_modal = '.main-inner'
                del button.attrs['up-target']

            content.append(button)

        return super().to_html(
            *content,
            '<div class="mdc-elevation-overlay"></div>',
            '<div class="mdc-button__ripple"></div>',
            **context,
        )


class Main(Main):
    pass


class ModalClose(Component):
    '''
    Close the modal when inserted by Unpoly.

    This should be the default behaviour of Unpoly, as documented, and is the
    case with modals that do not redirect on the same page, but for modals with
    a form with a ?_next= redirection then surprisingly unpoly does update the
    body but does not close the modal by default.

    This tag registers a compiler that calls up.modal.close() when it
    sees itself, given that unpoly does not execute scripts in
    response that it loads. As such, this tag MUST be in the initial body, but
    outside the elements that you will load in modal.

    Remove this hack when we figure the problem in unpoly.

    Equivalent of::

        Script(
            'up.compiler(".closemodal", function() { up.modal.close(); })',
            cls='closemodal',
        ),
    '''
    attrs = dict(cls='closemodal')

    def py2js():
        up.compiler('.closemodal', lambda: up.modal.close())


class Body(Body):
    style = 'margin: 0'

    def __init__(self, *content, **attrs):
        self.drawer = mdcDrawer(id='drawer')
        self.bar = mdcTopAppBar()
        self.main = Main(
            Div(
                *content,
                cls='main-inner',
            ),
            ModalClose(),
            cls='main mdc-drawer-app-content',
            id='main',
        )
        self.debug = settings.DEBUG
        super().__init__(
            self.bar,
            Div(
                self.drawer,
                self.main,
                cls='mdc-top-app-bar--fixed-adjust',
            ),
        )

    def py2js(self):
        up.compiler(
            '[data-mdc-auto-init]',
            lambda el: mdc.autoInit(el.parentElement)
        )
        if self.debug:
            up.log.enable()


class SpinnerOverlay(Div):
    sass = '''
    .SpinnerOverlay
        position: fixed
        width: 100%
        height: 100%
        top: 0
        left: 0
        right: 0
        bottom: 0
        background-color: rgba(255, 255, 255, 0.66)
        z-index: 20000
        cursor: pointer
        padding-top: 30vh

        .loader
            color: var(--mdc-theme-primary)
            opacity: 50%
            font-size: 90px
            text-indent: -9999em
            overflow: hidden
            width: 1em
            height: 1em
            border-radius: 50%
            margin: 72px auto
            position: relative
            -webkit-transform: translateZ(0)
            -ms-transform: translateZ(0)
            transform: translateZ(0)
            -webkit-animation: load6 1.7s infinite ease, round 1.7s infinite ease
            animation: load6 1.7s infinite ease, round 1.7s infinite ease

        @-webkit-keyframes load6
            0%
                box-shadow: 0 -0.83em 0 -0.4em, 0 -0.83em 0 -0.42em, 0 -0.83em 0 -0.44em, 0 -0.83em 0 -0.46em, 0 -0.83em 0 -0.477em
            5%, 95%
                box-shadow: 0 -0.83em 0 -0.4em, 0 -0.83em 0 -0.42em, 0 -0.83em 0 -0.44em, 0 -0.83em 0 -0.46em, 0 -0.83em 0 -0.477em
            10%, 59%
                box-shadow: 0 -0.83em 0 -0.4em, -0.087em -0.825em 0 -0.42em, -0.173em -0.812em 0 -0.44em, -0.256em -0.789em 0 -0.46em, -0.297em -0.775em 0 -0.477em
            20%
                box-shadow: 0 -0.83em 0 -0.4em, -0.338em -0.758em 0 -0.42em, -0.555em -0.617em 0 -0.44em, -0.671em -0.488em 0 -0.46em, -0.749em -0.34em 0 -0.477em
            38%
                box-shadow: 0 -0.83em 0 -0.4em, -0.377em -0.74em 0 -0.42em, -0.645em -0.522em 0 -0.44em, -0.775em -0.297em 0 -0.46em, -0.82em -0.09em 0 -0.477em
            100%
                box-shadow: 0 -0.83em 0 -0.4em, 0 -0.83em 0 -0.42em, 0 -0.83em 0 -0.44em, 0 -0.83em 0 -0.46em, 0 -0.83em 0 -0.477em
        @keyframes load6
            0%
                box-shadow: 0 -0.83em 0 -0.4em, 0 -0.83em 0 -0.42em, 0 -0.83em 0 -0.44em, 0 -0.83em 0 -0.46em, 0 -0.83em 0 -0.477em
            5%, 95%
                box-shadow: 0 -0.83em 0 -0.4em, 0 -0.83em 0 -0.42em, 0 -0.83em 0 -0.44em, 0 -0.83em 0 -0.46em, 0 -0.83em 0 -0.477em
            10%, 59%
                box-shadow: 0 -0.83em 0 -0.4em, -0.087em -0.825em 0 -0.42em, -0.173em -0.812em 0 -0.44em, -0.256em -0.789em 0 -0.46em, -0.297em -0.775em 0 -0.477em
            20%
                box-shadow: 0 -0.83em 0 -0.4em, -0.338em -0.758em 0 -0.42em, -0.555em -0.617em 0 -0.44em, -0.671em -0.488em 0 -0.46em, -0.749em -0.34em 0 -0.477em
            38%
                box-shadow: 0 -0.83em 0 -0.4em, -0.377em -0.74em 0 -0.42em, -0.645em -0.522em 0 -0.44em, -0.775em -0.297em 0 -0.46em, -0.82em -0.09em 0 -0.477em
            100%
                box-shadow: 0 -0.83em 0 -0.4em, 0 -0.83em 0 -0.42em, 0 -0.83em 0 -0.44em, 0 -0.83em 0 -0.46em, 0 -0.83em 0 -0.477em
        @-webkit-keyframes round
            0%
                -webkit-transform: rotate(0deg)
                transform: rotate(0deg)
            100%
                -webkit-transform: rotate(360deg)
                transform: rotate(360deg)
        @keyframes round
            0%
                -webkit-transform: rotate(0deg)
                transform: rotate(0deg)
            100%
                -webkit-transform: rotate(360deg)
                transform: rotate(360deg)
    '''  # noqa

    def py2js(self):
        up.proxy.config.slowDelay = 25
        up.compiler('.SpinnerOverlay', lambda element: [
            up.on('up:proxy:slow', lambda: up.element.show(element)),
            up.on('up:proxy:recover', lambda: up.element.hide(element)),
        ])

    def __init__(self):
        super().__init__(Div('...', cls='loader'), style='display:none')


class App(Html):
    body_class = Body
    scripts = [
        'https://unpkg.com/unpoly@1.0.0/dist/unpoly.js',
        # 'https://unpkg.com/unpoly@2.0.0-rc9/unpoly.min.js',
        # 'https://unpkg.com/unpoly@2.0.0-rc9/unpoly-migrate.js',
    ]
    stylesheets = [
        'https://unpkg.com/unpoly@1.0.0/dist/unpoly.css',
        # 'https://unpkg.com/unpoly@2.0.0-rc9/unpoly.min.css',
    ]
    sass = '''
    .up-modal .up-modal-viewport
        padding-left: 0
        padding-right: 0
        padding-top: 10px
    '''

    def to_html(self, *content, **context):
        if title := getattr(context['view'], 'title', None):
            self.head.content.append(Title(title))
        return super().to_html(*content, SpinnerOverlay(), **context)


class NarrowCard(Div):
    style = {
        'max-width': '95vw',
        'margin': 'auto',
        'padding': '1em',
        'margin-top': '2em',
        'button': {
            'width': '100%',
        },
        '.MDCField label': {
            'width': '100%',
        }
    }


@template('crudlfap/form.html', App)
@template('crudlfap/update.html', App)
@template('crudlfap/create.html', App)
class FormTemplate(FormContainer):
    def to_html(self, view, form, **context):
        back = ''
        if next_ := view.request.GET.get('next', ''):
            back = A(
                MDCButton(
                    _('Back'),
                    tag='span',
                ),
                href=next_,
            )
        return super().to_html(
            H3(view.title),
            Form(
                form,
                CSRFInput(view.request),
                back,
                MDCButtonRaised(getattr(view, 'title_submit', _('Submit'))),
            ),
        )


@template('crudlfap/api.html', Html)
class Swagger(Div):
    def py2js(self):
        el = document.getElementById('swagger-ui')
        ui = SwaggerUIBundle({
            url: el.attributes['schema-url'].value,
            dom_id: '#swagger-ui',
            deepLinking: true,
            presets: [
                SwaggerUIBundle.presets.apis,
                SwaggerUIStandalonePreset,
            ],
            plugins: [
                SwaggerUIBundle.plugins.DownloadUrl
            ],
            layout: "StandaloneLayout"
        })
        window.ui = ui

    def to_html(self, **context):
        registry = context['view'].registry
        return super().to_html(
            Script(src='/static/swagger-ui-bundle.js'),
            Script(src='/static/swagger-ui-standalone-preset.js'),
            Stylesheet(href='/static/swagger-ui.css'),
            Div(
                id='swagger-ui',
                schema_url=registry.views['schema'].url,
            ),
            **context
        )


@template('crudlfap/home.html', App, Container)
class Home(Div):
    def to_html(self, **context):
        site = Site.objects.get_current()
        return super().to_html(
            H1('Welcome to ' + site.name),
            MDCButtonRaised(
                'Login to continue',
                href=reverse('login'),
                up_target=UNPOLY_TARGET_ALL,
                tag='a',
            ),
            Div('Then, navigate with the menu button at the north west'),
            **context
        )


@template('registration/logged_out.html', App, NarrowCard)
class LoggedOut(Div):
    def to_html(self, **context):
        return super().to_html(
            H1(_('Log out')),
            P(_('Thanks for spending some quality time with the Web site today.')),  # noqa
            A(
                _('Log in again'),
                href=reverse('login'),
            ),
            **context,
        )


@template('crudlfap/detail.html', App)
class ObjectDetail(Div):
    def to_html(self, *content, **context):
        table = MDCDataTable()
        table.thead.attrs.style.display = 'none'
        table.table.attrs.style.width = '100%'
        table.attrs.data_mdc_auto_init = False

        for field in context['view'].display_fields:
            table.tbody.addchild(MDCDataTableTr(
                MDCDataTableTh(field['field'].verbose_name.capitalize()),
                MDCDataTableTd(field['value']),
            ))

        return super().to_html(
            PageMenu(),
            *content,
            NarrowCard(table),
            **context
        )

    def context(self, *content, **context):
        context['page-menu'] = context['view'].router.get_menu(
            'object',
            context['view'].request,
            object=context['view'].object,
        )
        return super().context(*content, **context)


class ListAction(Div):
    style = {'display': 'inline-block'}

    def onclick(element):
        link = element.attributes.href.value + '?'
        for checkbox in document.querySelectorAll('[data-pk]:checked'):
            link += '&pk=' + checkbox.attributes['data-pk'].value
        up.modal.visit(link, {target: '.main-inner'})


class ListActions(Component):
    tag = 'list-actions'
    style = {'display': 'none'}

    class HTMLElement:
        def connectedCallback(self):
            this.previousElementSibling.addEventListener(
                'change',
                this.change.bind(this),
            )

            this.change({
                'target': this.previousElementSibling.querySelector(':checked')
            })

        def change(self, event):
            if event.target and event.target.checked:
                this.style.display = 'block'
            elif not this.previousElementSibling.querySelector(':checked'):
                this.style.display = 'none'


@template('crudlfap/list.html', App)
class ObjectList(Div):
    def context(self, *content, **context):
        context['page-menu'] = context['view'].router.get_menu(
            'model',
            context['view'].request,
        )
        return super().context(*content, **context)

    def to_html(self, *content, **context):
        checkbox = None
        context['listactions'] = context['view'].router.get_menu(
            'list_action',
            context['view'].request,
        )

        if context['listactions']:
            table_checkbox = MDCCheckboxInput()
            table_checkbox.attrs.addcls = 'mdc-data-table__header-row-checkbox'

            checkbox = MDCDataTableTh(
                table_checkbox,
                addcls='mdc-data-table__header-cell--checkbox',
            )
        thead = MDCDataTableThead(tr=MDCDataTableHeaderTr(checkbox))

        for column in context['view'].table.columns:
            thead.tr.addchild(self.th_component(column, **context))

        # align "actions" title to the right with the buttons
        thead.tr.content[-1].attrs.style['text-align'] = 'right'

        table = MDCDataTableResponsive(thead=thead, style={
            'min-width': '100%',
            'border-width': 0,
        })

        for row in context['view'].table.paginated_rows:
            table.tbody.addchild(
                self.row_component(row, **context)
            )

        if context['listactions']:
            table.addchild(self.listactions_component(**context))

        table.addchild(self.pagination_component(**context))

        return super().to_html(
            *content,
            self.drawer_component(**context) or '',
            PageMenu(_next=context['view'].request.path),
            Div(
                self.search_component(**context) or '',
                table,
                cls='mdc-drawer__content',
                style='overflow-y: unset',
            ),
            **context,
        )

    def listactions_component(self, **context):
        return ListActions(
            *[
                ListAction(
                    MDCButton(
                        view.label,
                        icon=view.icon,
                        style='color: ' + getattr(view, 'color', 'inherit'),
                    ),
                    title=view.title,
                    href=view.url,
                    up_modal='.main-inner',
                    up_target=False,
                ) for view in context['view'].listactions
            ],
        )

    def search_component(self, **context):
        search_form = self.search_form_component(**context)

        filterset = getattr(context['view'], 'filterset', None)
        if not filterset:
            return search_form

        toggle = MDCDrawerToggle(
            Button(
                'filter_list',
                cls='material-icons mdc-icon-button',
                style='--mdc-ripple-fg-size:28px; --mdc-ripple-top:10px;',
            ),
            data_drawer_id='page-drawer',
        )

        filters_chips = Div(
            toggle=toggle if context['view'].filterset.form.fields else '',
            search=search_form or '',
            chips=Div(
                cls='mdc-chip-set',
                role='grid',
                style='display: inline-block',
            )
        )

        def remove_filter_url(name):
            get = context['view'].request.GET.copy()
            if name in get:
                del get[name]
            return context['view'].request.path_info + '?' + get.urlencode()

        for bf in filterset.form.visible_fields():
            value = context['view'].request.GET.get(bf.name, '')
            if not value:
                continue
            chip = MDCChip(
                Span(
                    role='button',
                    tabindex='0',
                    cls='mdc-chip__primary-action',
                    text=Span(
                        bf.label,
                        ': ',
                        str(bf.form.cleaned_data[bf.name]),
                        cls='mdc-chip__text',
                    ),
                ),
                icon=I(
                    'cancel',
                    cls=(
                        'material-icons',
                        'mdc-chip__icon',
                        'mdc-chip__icon--trailing',
                    ),
                    tabindex='-1',
                    role='button',
                ),
                tag='a',
                href=remove_filter_url(bf.name),
                up_target='main',
            )
            filters_chips.chips.addchild(chip)
        return filters_chips

    def drawer_component(self, **context):
        filterset = context['view'].filterset
        if not filterset.form.fields:
            return
        form = Form(
            method='get',
            action=context['view'].request.path_info,
            up_autosubmit=True,
            up_delay='200ms',
            up_target='.mdc-drawer__content',
        )
        for key in ('sort', 'q'):
            if key in context['view'].request.GET:
                form.addchild(Input(
                    name=key,
                    value=context['view'].request.GET[key],
                    type='hidden'
                ))
        for bf in filterset.form.visible_fields():
            filterfield = MDCFilterField(
                label=bf.label,
                widget=bf.to_component()
            )
            value = context['view'].request.GET.get(bf.name, '')
            if not value:
                filterfield.widget.style.display = 'none'
            form.addchild(filterfield)
        drawer = Aside(
            MDCDrawerToggle(
                MDCButton(_('Close'), icon='close'),
                data_drawer_id='page-drawer',
                style='text-align: right',
            ),
            Div(
                form,
                cls='mdc-drawer-app-content',
            ),
            id='page-drawer',
            cls='mdc-drawer mdc-drawer--dismissible',
            style='padding: 1em',
        )
        return drawer

    def search_form_component(self, **context):
        search_form = getattr(context['view'], 'search_form', None)

        if not search_form:
            return

        search_form = InlineForm(
            search_form,
            method='get',
            action=context['view'].request.path_info,
            up_autosubmit=True,
            up_delay='200ms',
            up_target='.mdc-data-table, .mdc-chip-set',
            style='margin-left: 14px',
        )
        for k, v in context['view'].request.GET.items():
            if k == 'q':
                continue
            search_form.addchild(Input(
                name=k,
                value=v,
                type='hidden',
            ))
        return search_form

    def row_component(self, row, **context):
        show_checkbox = False
        for listaction in context['listactions']:
            if listaction.clone(object=row.record)().has_perm():
                show_checkbox = True

        if show_checkbox:
            checkboxinput = MDCCheckboxInput(
                data_pk=str(row.record.pk)
            )
            checkboxinput.attrs.addcls = 'mdc-data-table__row-checkbox'
            tr = MDCDataTableTr(
                MDCDataTableTd(
                    checkboxinput,
                    addcls='mdc-data-table__cell--checkbox',
                )
            )
        elif context['listactions']:
            tr = MDCDataTableTr(MDCDataTableTd())
        else:
            tr = MDCDataTableTr()

        for column, cell in row.items():
            # todo: localize values
            tr.addchild(
                MDCDataTableTd(cell, data_label=column.header),
            )
            # todo: if is numeric
            # td.attrs.addcls = 'mdc-data-table__header-cell--numeric'
        return tr

    def th_component(self, column, **context):
        th = MDCDataTableTh(
            wrapper=Div(
                cls='mdc-data-table__header-cell-wrapper',
                label=Div(
                    cls='mdc-data-table__header-cell-label',
                    style='font-weight: 500',
                    text=Text(column.header),
                ),
            )
        )

        # sorting
        if column.orderable:
            th.attrs.addcls = 'mdc-data-table__header-cell--with-sort'
            if column.is_ordered:
                th.attrs.addcls = 'mdc-data-table__header-cell--sorted'
            get = context['view'].request.GET.copy()
            get['sort'] = column.order_by_alias.next
            href = ''.join([
                context['view'].request.path_info,
                '?',
                get.urlencode(),
            ])
            th.wrapper.content += [
                A(
                    cls=(
                        'mdc-icon-button',
                        'material-icons',
                        'mdc-data-table__sort-icon-button',
                    ),
                    aria_label='Sort by dessert',
                    aria_describedby='dessert-status-label',
                    up_target='table',
                    href=href,
                    text=Text(
                        'arrow_upward'
                        if column.order_by_alias.is_descending
                        else 'arrow_downward'
                    ),
                ),
                Div(
                    cls='mdc-data-table__sort-status-label',
                    aria_hidden='true',
                    id='dessert-status-label',
                ),
            ]
        return th

    def pagination_component(self, **context):
        def pageurl(n):
            get = context['view'].request.GET.copy()
            get['page'] = n
            return context['view'].request.path_info + '?' + get.urlencode()

        page = context['view'].table.page
        navigation = Div(
            Div(
                cls='mdc-data-table__pagination-total',
                text=Text(''.join([
                    str(page.start_index()),
                    '-',
                    str(page.paginator.per_page * page.number),
                    ' / ',
                    str(page.paginator.count),
                ]))
            ),
            A(
                cls=(
                    'mdc-icon-button',
                    'material-icons',
                    'mdc-data-table__pagination-button',
                ),
                disabled=page.number == 1,
                href=pageurl(1),
                icon=Div(cls='mdc-button__icon', text=Text('first_page')),
                up_target='.mdc-data-table',
            ),
            A(
                cls=(
                    'mdc-icon-button',
                    'material-icons',
                    'mdc-data-table__pagination-button',
                ),
                disabled=not page.has_previous(),
                icon=Div(cls='mdc-button__icon', text=Text('chevron_left')),
                href=pageurl(
                    page.number - 1
                    if page.has_previous()
                    else 1
                ),
                up_target='.mdc-data-table',
            ),
            A(
                cls=(
                    'mdc-icon-button',
                    'material-icons',
                    'mdc-data-table__pagination-button',
                ),
                disabled=not page.has_next(),
                icon=Div(cls='mdc-button__icon', text=Text('chevron_right')),
                href=pageurl(
                    page.number + 1
                    if page.has_next()
                    else page.paginator.num_pages
                ),
                up_target='.mdc-data-table',
            ),
            A(
                cls=(
                    'mdc-icon-button',
                    'material-icons',
                    'mdc-data-table__pagination-button',
                ),
                disabled=page.paginator.num_pages == page.number,
                icon=Div(cls='mdc-button__icon', text=Text('last_page')),
                href=pageurl(page.paginator.num_pages),
                up_target='.mdc-data-table',
            ),
            cls='mdc-data-table__pagination-navigation',
        )
        perpage = Div(
            Div(
                _('Rows per page'),
                cls='mdc-data-table__pagination-rows-per-page-label'
            ),
            select=MDCSelectPerPage(
                addcls=(
                    'mdc-select--outlined',
                    'mdc-select--no-label',
                    'mdc-data-table__pagination-rows-per-page-select',
                ),
                select=Select(*[
                    Option(
                        str(i),
                        value=i,
                        selected=page.paginator.per_page == i
                    )
                    for i in (3, 5, 7, 10, 25, 100)
                ])
            ),
            cls='mdc-data-table__pagination-rows-per-page',
        )
        return MDCDataTablePagination(
            perpage=perpage,
            navigation=navigation,
        )


class mdcTopAppBar(Header):
    def __init__(self, title='', buttons=None):
        self.title = title
        self.buttons = buttons or []
        super().__init__(
            Div(cls='mdc-top-app-bar__row'),
            cls='mdc-top-app-bar app-bar',
            id='app-bar',
            data_mdc_auto_init='MDCTopAppBar',
        )

    def to_html(self, view, **context):
        self.content[0].content = [Component(
            Section(
                Button(
                    'menu',
                    cls='material-icons mdc-top-app-bar__navigation-icon mdc-icon-button',  # noqa
                ),
                Span(
                    getattr(view, 'title', getattr(settings, 'SITE_NAME', '')),
                    cls='mdc-top-app-bar__title',
                ),
                cls='mdc-top-app-bar__section mdc-top-app-bar__section--align-start',  # noqa
            ),
            cls='mdc-top-app-bar__section mdc-top-app-bar__section--align-start',  # noqa
            tag='section',
        )]
        return super().to_html(**context)

    def nav():
        window.drawer.open = not window.drawer.open

    def close():
        if window.visualViewport.width < 800:
            window.drawer.open = False

    def py2js(self):
        up.compiler('#drawer, #app-bar, #main', self.setup)

    def setup():
        drawer = document.getElementById('drawer')
        app_bar = document.getElementById('app-bar')
        main = document.getElementById('main')

        window.drawer = mdc.drawer.MDCDrawer.attachTo(drawer)
        topAppBar = mdc.topAppBar.MDCTopAppBar.attachTo(app_bar)
        topAppBar.setScrollTarget(main)
        topAppBar.listen('MDCTopAppBar:nav', self.nav)

        for link in drawer.querySelectorAll('a'):
            link.addEventListener('click', self.close)


class mdcDrawer(Aside):
    menu_hooks = []

    def __init__(self, *content, **attrs):
        super().__init__(
            Div(
                *content,
                cls='mdc-drawer__content',
            ),
            cls='mdc-drawer mdc-drawer--dismissible mdc-top-app-bar--fixed-adjust',  # noqa
            data_mdc_auto_init='MDCDrawer',
            **attrs,
        )

    def to_html(self, *content, view, **context):
        request = view.request
        from .site import site

        menu_content = []
        if request.user.is_authenticated:
            menu_content.append(MDCListItem(str(request.user.email)))
        for view in site.get_menu('main', request):
            icon = getattr(view, 'icon', None)
            title = getattr(view, 'title', '')
            attrs = dict()
            if getattr(view, 'unpoly', True):
                attrs['up_target'] = UNPOLY_TARGET_ALL

            menu_content.append(
                MDCListItem(
                    title.capitalize(), icon=icon,
                    href=view.url,
                    style='text-decoration: none',
                    tag='a',
                    **attrs,
                ),
            )
        for hook in self.menu_hooks:
            menu_content = hook(request, menu_content)

        content = menu_content

        if request.session.get('become_user', None):
            content.append(
                A(
                    ' '.join([
                        str(_('Back to your account')),
                        request.session['become_user_realname'],
                    ]),
                    href=reverse('crudlfap:su'),
                    up_target=UNPOLY_TARGET_ALL,
                )
            )

        return super().to_html(MDCList(*content))


class mdcAppContent(Div):
    def __init__(self, *content):
        super().__init__(
            Component(
                *content,
                tag='main',
                cls='main-content',
                id='main-content',
            ),
            cls='mdc-drawer-app-content mdc-top-app-bar--fixed-adjust',
        )


class mdcSwitch(Component):
    def __init__(
        self, name, label=None, value=None, type=None, errors=None, help=None,
        required=False,
    ):
        super().__init__(
            Div(
                Div(cls='mdc-switch__track'),
                Div(
                    Div(cls='mdc-switch__thumb'),
                    Input(
                        type='checkbox',
                        id='id_' + name,
                        role='switch',
                        aria_checked='true' if bool(value) else '',
                        cls='mdc-switch__native-control',
                        checked=bool(value)
                    ),
                    cls='mdc-switch__thumb-underlay',
                ),
                cls='mdc-switch mdc-switch--checked',
                data_mdc_auto_init='MDCSwitch',
            ),
            Label(
                label or name.capitalize(),
                **{'for': 'id_' + name},
            )
        )



"""CRUD :eMixins that can be used in Actions or Views."""
import copy
import json

from django import forms
from django import http
from django.contrib.admin.models import ADDITION, CHANGE, DELETION, LogEntry
from django.contrib.contenttypes.models import ContentType

from crudlfap import html


class CreateMixin:
    style = 'success'
    icon = 'add'
    default_template_name = 'crudlfap/create.html'
    template_name_suffixes = ['create', 'form']
    controller = 'modal'
    action = 'click->modal#open'
    color = 'green'
    log_action_flag = ADDITION
    menus = ['model']
    permission_shortcode = 'add'

    def form_valid(self):
        self.object = self.form.save()
        return super().form_valid()


class ActionMixin:
    def has_perm(self):
        """
        Call has_perm_object or return True.

        When called without an object, return True if router agrees.

        Otherwise, return the result of has_perm_object(), that you must
        implement, to return wether permission is accepted for a particular
        object.

        To override yourself, again go inside an if super().has_perm(): to
        benefit from this behaviour.
        """
        if super().has_perm():
            if hasattr(self, 'object'):
                return self.has_perm_object()
            return True

    def has_perm_object(self):
        """
        Override this method: test self.object then return True.

        By default, return True.
        """
        return True


class DeleteMixin(ActionMixin):
    style = 'danger'
    fa_icon = 'trash'
    icon = 'delete'
    success_url_next = True
    color = 'red'
    log_action_flag = DELETION
    controller = 'modal'
    action = 'click->modal#open'
    form_class = forms.Form
    permission_shortcode = 'delete'

    def form_valid(self):
        if hasattr(self, 'object_list'):
            self.result = copy.copy(self.object_list).delete()
        else:
            self.result = self.object.delete()
        return super().form_valid()

    def get_success_url(self):
        return self.request.GET.get('_next', self.router['list'].reverse())


class DetailMixin:
    fa_icon = 'search-plus'
    icon = 'search'
    default_template_name = 'crudlfap/detail.html'
    color = 'blue'
    menus_display = ['object', 'object_detail']

    @classmethod
    def get_urlpath(cls):
        """Identify the object by slug or pk in the pattern."""
        return r'<{}>'.format(cls.urlfield)

    def get_JSONField_display(self, name):  # noqa
        value = getattr(self.object, name, "")
        formated = json.dumps(value, indent=4)
        return f'<pre>{formated}</pre>'

    def get_title(self):
        return str(self.object)

    def get_visible_fields(self):
        return [
            f.name for f in self.model._meta.fields
            if self.fields == '__all__' or f.name not in self.exclude
        ]

    def get_display_fields(self):
        """Table field rendering"""
        self.display_fields = [
            {
                'field': self.model._meta.get_field(field),
                'value': self.get_field_display(field),
                'name': field,
            }
            for field in self.visible_fields
        ]

    def get_field_display(self, name):
        value_getter = '_'.join(['get', name, 'display'])
        if hasattr(self.object, value_getter):
            return getattr(self.object, value_getter)()
        if hasattr(self, value_getter):
            return getattr(self, value_getter)()
        type_getter = '_'.join([
            'get',
            type(self.model._meta.get_field(name)).__name__,
            'display',
        ])
        if hasattr(self, type_getter):
            return getattr(self, type_getter)(name)
        value = getattr(self.object, name)
        if hasattr(value, 'get_absolute_url'):
            return html.A(
                str(value),
                href=value.get_absolute_url(),
            ).render()
        return value

    def get_json_fields(self):
        return self.visible_fields

    def get_FIELD_json(self, obj, field):
        value = getattr(obj, field)
        if self.router and type(value) in self.router.registry:
            value = self.router.registry[type(value)].serialize(obj)
        elif value and not isinstance(value, (str, int, float, bool)):
            value = str(value)
        return value

    def serialize(self):
        if self.router:
            return self.router.serialize(self.object, self.json_fields)
        return {
            field: getattr(
                self,
                f'get_{field}_json',
                'get_FIELD_json',
            )(self.object, field)
            for field in self.json_fields
        }

    def json_get(self, request, *args, **kwargs):
        return http.JsonResponse(self.serialize())


class HistoryMixin:
    icon = 'history'
    template_name_suffix = '_history'
    default_template_name = 'crudlfap/history.html'
    controller = None
    action = None

    def get_object_list(self):
        ctype = ContentType.objects.get_for_model(self.model)
        return LogEntry.objects.filter(
            content_type=ctype,
            object_id=self.object.pk,
        )


class ListMixin:
    default_template_name = 'crudlfap/list.html'
    body_class = 'full-width'
    menus = ['main', 'model']
    title_heading = None

    def get_icon(self):
        if self.router:
            return self.router.icon
        return 'list'

    def get_title(self):
        return self.model._meta.verbose_name_plural.capitalize()

    def get_urlpath(self):
        return ''

    def get_title_heading(self):
        return self.model._meta.verbose_name_plural.capitalize()

    def get_swagger_get(self):
        '''TODO
        parameters = {
                    'collectionFormat': 'multi',
                    'description': 'Status values to filter',
                    'in': 'query',
                    'items': {
                        'default': 'available',
                        'enum': ['available', 'pending', 'sold'],
                        'type': 'string'
                    },
                    'name': 'status',
                    'required': True,
                    'type': 'array'
                }
        '''
        return {
            # 'description': self.title,
            # 'operationId': 'findPetsByStatus',
            'parameters': [],
            'produces': ['application/json'],
            'responses': {
                '200': {
                    'description': 'successful operation',
                    'schema': {
                        'items': {
                            '$ref': '#/definitions/' + self.model.__name__
                        },
                        'type': 'array'
                    }
                },
                '400': {'description': 'Invalid status value'}
            },
            'summary': self.title,
            'tags': self.swagger_tags
        }


class UpdateMixin:
    icon = 'edit'
    default_template_name = 'crudlfap/update.html'
    template_name_suffixes = ['create', 'form']
    controller = 'modal'
    action = 'click->modal#open'
    color = 'orange'
    locks = True
    log_action_flag = CHANGE
    permission_shortcode = 'change'

    def get_form_fields(self):
        if hasattr(self, 'update_fields'):
            return self.update_fields
        if hasattr(self.router, 'update_fields'):
            return self.router.update_fields
        return super().get_form_fields()

    def form_valid(self):
        self.object = self.form.save()
        return super().form_valid()



"""
Test security with crudlfap_example.song setup.

It's a bit paranoid, but i'll sleep better with this: i don't trust users, i
don't trust myself either :)
"""

import pytest

from crudlfap import shortcuts as crudlfap
from crudlfap_auth.crudlfap import User
from crudlfap_example.artist.models import Artist
from crudlfap_example.song.models import Song


def user(**attrs):
    user, created = User.objects.get_or_create(**attrs)
    if created:
        user.set_password('password')
        user.save()
    return user


def user_client(client, **userattrs):
    if not userattrs.get('username', None):
        return client
    user(**userattrs)
    if not client.login(username=userattrs['username'], password='password'):
        raise Exception('Could not auth as ' + userattrs['username'])
    return client


@pytest.fixture
def song0():
    user0 = user(username='user0')
    artist = Artist.objects.get_or_create(name='artist0')[0]
    return Song.objects.get_or_create(
        artist=artist, name='song0', owner=user0)[0]


cases = [
    (dict(), False),
    (dict(username='user0'), True),
    (dict(username='user1'), False),
    (dict(username='staff', is_staff=True), True),
    (dict(username='superuser', is_superuser=True), True),
]


@pytest.mark.parametrize('userattrs,expected', cases)
@pytest.mark.parametrize('url', ['detail', 'update', 'delete'])
@pytest.mark.django_db
def test_object_views_object_for_user(client, userattrs, expected, url, song0):
    view = crudlfap.site[Song][url].clone(object=song0)
    url = view.url
    res = user_client(client, **userattrs).get(url)
    assert res.status_code == 200 if expected else 404


@pytest.mark.parametrize('userattrs,expected', cases)
@pytest.mark.django_db
def test_list_view_get_objects_for_user(client, userattrs, expected, song0):
    res = user_client(client, **userattrs).get(crudlfap.site[Song]['list'].url)

    assert res.status_code == 200 if userattrs else 302

    if expected:
        assert b'song0' in res.content
    else:
        assert b'song0' not in res.content



"""Django AppConfig for the crudlfap module."""

from django import apps
from django.conf import settings


def _installed(*apps):
    for app in apps:
        if app not in settings.INSTALLED_APPS:
            return False
    return True


class DefaultConfig(apps.AppConfig):
    """Default AppConfig."""

    name = 'crudlfap'

    def ready(self):
        super().ready()
        self.module.autodiscover()



import collections

import django_tables2 as tables
from django.db import models
from django.utils.safestring import mark_safe
from django.utils.translation import ugettext_lazy as _
from django_tables2.config import RequestConfig

from crudlfap import html


class UnpolyLinkColumn(tables.LinkColumn):
    def render(self, record, value):
        return super().render(record, value).replace(
            '<a ',
            '<a up-target="{html.A.attrs["up-target"]}"',
        )


class ActionsColumn(tables.Column):
    empty_values = ()

    def __init__(self, **kwargs):
        kwargs.setdefault('default', True)
        super().__init__(**kwargs)

    def render(self, record, table, value, **kwargs):
        from crudlfap.site import site
        out = html.Div(
            *site[type(record)].get_menu_component(
                'object',
                table.request,
                object=record
            ).content,
            style='display:flex;flex-direction:row-reverse',
        ).render()
        return mark_safe(out)


class Table(tables.Table):
    def before_render(self, request):
        self.request = request

    def render_tags(self, record):
        html = []
        for tag in record.tags.all():
            html.append(str(tag))

        return ' '.join(html)


class TableMixin(object):
    def get_table_fields(self):
        if self.table_sequence:
            self.table_fields = [
                f.name
                for f in self.model._meta.fields
                if f.name in self.table_sequence
            ]
        else:
            self.table_fields = [
                f.name
                for f in self.model._meta.fields
                if f.name not in self.exclude
            ]
        return self.table_fields

    def get_table_link_fields(self):
        if not hasattr(self.model, 'get_absolute_url'):
            return []

        for field in self.table_fields:
            model_field = self.model._meta.get_field(field)
            if isinstance(model_field, models.CharField):
                return [field]

        return []

    def get_table_meta_link_columns(self):
        return {i: UnpolyLinkColumn() for i in self.table_link_fields}

    def get_table_meta_action_columns(self):
        return dict(
            crudlfap=ActionsColumn(
                verbose_name=_('Actions'),
                orderable=False,
            )
        )

    def get_table_sequence(self):
        return None

    def get_table_columns(self):
        return dict()

    def get_table_meta_attributes(self):
        attrs = dict(model=self.model)

        if self.table_sequence:
            attrs['sequence'] = self.table_sequence

        if self.table_fields:
            attrs['fields'] = self.table_fields

        if self.listactions:
            if 'sequence' in attrs:
                attrs['sequence'] = list(attrs['sequence'])
                attrs['sequence'].append('crudlfap')
            else:
                attrs['sequence'] = ['...', 'crudlfap']

        return attrs

    def get_table_meta_class(self):
        return type('Meta', (object,), self.table_meta_attributes)

    def get_table_class_attributes(self):
        attrs = collections.OrderedDict(
            Meta=self.table_meta_class,
        )
        attrs.update(self.table_meta_link_columns)
        attrs.update(self.table_meta_action_columns)
        attrs.update(self.table_columns)
        return attrs

    def get_table_class(self):
        return Table

    def build_table_class(self):
        bases = (self.table_class,)
        if (self.table_class != Table
                and not issubclass(self.table_class, Table)):

            bases = (self.table_class, Table)

        return type(
            '{}Table'.format(self.model.__name__),
            bases,
            self.table_class_attributes
        )

    def get_table_kwargs(self):
        return dict(data=self.object_list)

    def get_max_per_page(self):
        return 100

    def get_table_pagination(self):
        if not self.paginate_by:
            return True
        per_page = int(self.request.GET.get('per_page', self.paginate_by))
        if per_page > self.max_per_page:
            return self.max_per_page
        return dict(per_page=per_page)

    def get_table(self):
        kwargs = self.table_kwargs
        self.table = self.build_table_class()(**kwargs)
        RequestConfig(
            self.request,
            paginate=self.table_pagination,
        ).configure(self.table)
        return self.table

    def get_paginate_by(self):
        return 10



from django.contrib.auth.models import Permission
from django.contrib.contenttypes.models import ContentType
from django.db import models
from lookupy import Collection

from crudlfap.site import site as default_site


class ControllerManager(models.Manager):
    def __init__(self, site=None):
        super().__init__()
        self.site = site or default_site

    def get_queryset(self):
        items = [
            Controller.factory(i) for i in self.site.values()
        ]
        return Collection(items)


class Controller(models.Model):
    router = models.Field()
    model = models.Field()
    app = models.Field()

    objects = ControllerManager()

    class Meta:
        managed = False

    def __str__(self):
        return self.pk

    @classmethod
    def factory(cls, router):
        return cls(
            pk=f'{router.model._meta.app_label}.{router.model.__name__}',
            router=router,
            app=str(router.model._meta.app_config.verbose_name),
            model=str(router.model._meta.verbose_name).capitalize(),
        )


class URLManager(models.Manager):
    def __init__(self, site=None):
        super().__init__()
        self.site = site or default_site

    def get_queryset(self):
        views = [
            URL.factory(i) for i in self.site.views
        ]

        for router in self.site.values():
            views += [
                URL.factory(i) for i in router.views
            ]

        return Collection(views)


class URL(models.Model):
    controller = models.ForeignKey(
        Controller,
        on_delete=models.SET_NULL,
        null=True,
    )
    urlpath = models.Field()
    fullurlpath = models.Field()
    urlname = models.Field()
    urlfullname = models.Field()
    view = models.Field()
    model = models.Field()

    objects = URLManager()

    class Meta:
        managed = False

    def __str__(self):
        return self.view.label

    @property
    def content_type(self):
        return ContentType.objects.get_for_model(self.view.model)

    @property
    def codename(self):
        if not self.view.model:
            return self.view.permission_shortcode
        return '_'.join((
            self.view.permission_shortcode,
            self.view.model.__name__.lower(),
        ))

    def get_or_create_permission(self):
        return Permission.objects.update_or_create(
            content_type=self.content_type,
            codename=self.codename,
            defaults=dict(
                name=getattr(
                    self.view,
                    'title_menu',
                    self.view.permission_shortcode
                ),
            )
        )[0]

    @classmethod
    def factory(cls, view):
        kwargs = dict(
            pk=f'{view.__name__}',
            view=view,
            urlpath=view.urlpath,
            urlname=view.urlname,
            urlfullname=view.urlfullname,
        )

        if view.model:
            kwargs['model'] = view.model

        if view.router:
            kwargs['controller'] = Controller.factory(view.router)
            kwargs['fullurlpath'] = '/'.join((
                view.router.registry.urlpath,
                view.router.urlpath,
                view.urlpath,
            ))

        url = cls(**kwargs)
        return url



# flake8: noqa: N805
"""
CRUDLFA+ introduces an MVC-ish pattern, as the Router class is meant to sit
between a Model class and its set of View. Your views will have to inherit from
Route to work in Router.views. This structural decision made for you by
CRUDLFA+ was not exactly designed: it's an open source rewrite of a module that
was ordered in a proprietary project.
"""
import re

from django import http
from django.urls import path, reverse, reverse_lazy
from django.utils.module_loading import import_string

from .factory import Factory, FactoryMetaclass
from .utils import guess_urlfield


class RouteMetaclass(FactoryMetaclass):
    """Base autocalculations for views.

    .. py:data:: app_name

        The view's app name.

    .. py:data:: model

        The view's model if any.

    .. py:data:: urlpath

        The path for the url path definition.

    .. py:data:: label

        The view label, serves as key in a Router.views.

    .. py:data:: urlpattern

        The Django URL path() instance, for inclusion in url lists.

    .. py:data:: urlfullname

        The full name to reverse the URL, with namespaces if any.

    .. py:data:: urlfield

        The default model field that will be use to match in the URL. It can be
        pk, or name, slug ...
    """
    router = None

    def get_app_name(cls):
        """Return the model's app_name or None."""
        return cls.model._meta.app_label if cls.model else None

    def get_model(cls):
        """Return the router's model or None."""
        return cls.router.model if cls.router else None

    def get_urlpath(cls):
        """Return the urlname."""
        return cls.urlname

    def get_urlname(cls):
        """Return a string that can be used as url name."""
        urlname = cls.__name__.lower()
        if urlname.endswith('view'):
            urlname = urlname[:-4]
        elif urlname.endswith('route'):
            urlname = urlname[:-5]

        if cls.model:
            model_name = cls.model._meta.model_name.lower()
            if urlname.startswith(model_name):
                urlname = urlname[len(model_name):]

        if not urlname and cls.model:
            urlname = cls.model._meta.model_name

        return urlname or None

    def get_label(cls):
        """Return a readable label for this view.

        Strips View and Route from class name, also removes the model class
        name if it finds it: for YourModelUpdateView this returns `update`.
        """
        name = re.sub('(View|Route)$', '', cls.__name__)
        if cls.model:
            name = re.sub('^' + cls.model.__name__, '', name)
        return re.sub("([a-z])([A-Z])","\g<1> \g<2>", name)

    def get_urlpattern(cls):
        """Return the Django URL object to include in a urlpatterns."""
        return path(cls.urlpath, cls.as_view(), name=cls.urlname)

    def get_urlfullname(cls):
        """Return the url name eventually with router and site namespaces."""
        if cls.router and cls.registry:
            return '{}:{}:{}'.format(
                cls.router.registry.app_name,
                cls.router.namespace,
                cls.urlname
            )
        elif cls.registry:
            return '{}:{}'.format(
                cls.registry.app_name,
                cls.urlname
            )
        elif cls.router:
            return '{}:{}'.format(
                cls.router.namespace,
                cls.urlname
            )
        else:
            return cls.urlname

    def get_urlfield(cls):
        """Return the router urlfield if any, else guess_urlfield()"""
        if cls.router and cls.router.urlfield:
            return cls.router.urlfield
        return guess_urlfield(cls.model)


class Route(Factory, metaclass=RouteMetaclass):
    """The mixin for Views that will make it compatible with Router.

    .. py:data:: authenticate

        False by default, it makes the default has_perm() implementation
        require Django permission.

    .. py:data:: urlargs

        Args that should be passed to reverse() along with Route.urlfullname.

    .. py:data:: url

        Absolute url to the view, relying on Route.urlfullname and
        Route.urlargs.

    You will be able to check if a user has access to a view with a given
    object for example as such::

        crudlfap.site[YourModel]['detail'].clone(
            request=request,
            object=obj,
        ).has_perm()

    If you want to open a View to all, set authenticate=False, examples::

        class YourDetailView(DetailView):
            authenticate = False

        class YourRouter(Router):
            views = [
                YourDetailView,
                ListView.clone(authenticate=False),  # example with clone
            ]

    Without authenticate=False, the default has_perm() implementation requires the
    request user to have the permission corresponding to the
    permission_fullcode attribute.

    To create the permission with permission_fullcode, you can browse in your
    CRUDLFA+ site and navigate to URL list view, for each URL you have link in
    the menu called "authorized" that lets you select which groups have this
    permission: it will auto-create the permission in the database if
    necessary.
    """

    @classmethod
    def reverse(cls, *args, **kwargs):
        """Reverse a url to this view with the given args."""
        return reverse_lazy(cls.urlfullname, args=args, kwargs=kwargs)

    def get_urlargs(self):
        """
        Return args for reversing this view url from self.
        See ``self.reverse()`` for detail.
        """
        return []

    def get_url(self):
        """
        Return the URL for this view given its current state.
        Given that the ``reverse()`` method is a class method, this should
        allow things like::
            url = YourView(object=your_object).url
        """
        return self.reverse(*self.urlargs)

    def get_permission_shortcode(self):
        """Return the middle part for the view permission.

        Returns the urlname by default.
        """
        return self.urlname

    def get_permission_codename(self):
        """Return the codename attribute for the view Permission."""
        if not self.model:
            return self.permission_shortcode
        return f'{self.permission_shortcode}_{self.model._meta.model_name}'

    def get_permission_fullcode(self):
        """
        Return a string with the app name, permission_shortcode and model name.
        """
        return f'{self.app_name}.{self.permission_codename}'

    def get_authenticate(self):
        return True

    def has_perm(self):
        """
        Called to decide if user has the permission to execute this view.

        This is what you should override to code view level permission logic,
        unless you want to use a Django permission backend.

        If unsure, you probably need to just define Router.has_perm.

        Default behaviour:

        - return True if the view ``authenticate`` attribute has been
          overridden to False
        - return self.router.has_perm(self) if the view has a router
        - return self.has_perm_backend otherwise, to let a Django permission
          backend decide
        """
        if not self.authenticate:
            return True

        if self.router:
            return self.router.has_perm(self)

        return self.has_perm_backend()

    def has_perm_backend(self):
        """
        Django Authentication backend has_perm() call.

        This is called by Route.has_perm by default, from Router.has_perm is
        the view has a router object, and useful only if you implement your own
        permission backend, or add crudlfap_auth.backends.ViewBackend to
        settings.AUTHENTICATION_BACKENDS, or implement your own View permission
        backend.

        If unsure, you probably need to just define Router.has_perm.
        """
        return self.request.user.has_perm(self.permission_fullcode, self)

    def get_registry(self):
        if self.router:
            return self.router.registry
        from .site import site
        return site

    def get_allowed_groups(self):
        if not self.router:
            return []
        return getattr(self.router, 'allowed_groups', [])

    def dispatch(self, request, *args, **kwargs):
        """This will run has_perm prior to super().dispatch()."""
        if not self.has_perm():
            if not request.user.is_authenticated:
                return http.HttpResponseRedirect(
                    reverse('login') + '?next=' + request.path_info
                )
            else:
                return http.HttpResponseForbidden()
        HTTP_ACCEPT = request.META.get('HTTP_ACCEPT', '')
        json_handler = (
            request.content_type == 'application/json'
            or HTTP_ACCEPT.startswith('application/json')
        )
        if json_handler and request.method.lower() in self.http_method_names:
            handler = f'json_{self.request.method.lower()}'
            if hasattr(self, handler):
                result = getattr(self, handler)(request, *args, **kwargs)
                if isinstance(result, dict):
                    return http.JsonResponse(result)
                return result
        return super().dispatch(request, *args, **kwargs)

    @classmethod
    def factory(cls, view, **attributes):
        if isinstance(view, str):
            view = import_string(view)
        return type(view.__name__, (view, cls), attributes)

    @classmethod
    def abstract(cls, **kwargs):
        """Return an instance of this view."""
        return cls(**kwargs)

    def get_swagger_path_definition(self):
        result = dict()
        for method in self.http_method_names:
            method_def = getattr(self, f'swagger_{method}', None)
            if not method_def:
                continue
            result[method] = method_def
        return result



import logging

from django.conf import settings
from django.template.response import TemplateResponse
from django.utils.translation import ugettext as _

logger = logging.getLogger()


class TemplateMixin:
    """
    Replacement for Django's TemplateResponseMixin.

    This allows to configure "last resort" templates for each class, and thus
    to provide a working CRUD out of the box.
    """

    def get(self, request, *args, **kwargs):
        return self.render_to_response()

    def render_to_response(self):
        """
        Return a response, using the `response_class` for this view, with a
        template rendered with the given context.

        Pass response_kwargs to the constructor of the response class.
        """
        return self.response_class(
            request=self.request,
            template=self.template_names,
            context=self.context,
            using=self.template_engine,
            **self.response_kwargs
        )

    def get_template_engine(self):
        return getattr(
            settings, 'CRUDLFAP', {}
        ).get('TEMPLATE_ENGINE', 'ryzom')

    def get_response_class(self):
        return TemplateResponse

    def get_response_kwargs(self):
        return dict(content_type='text/html')

    def get_context(self, **context):
        context['view'] = self
        self.context = context

    def get_view_label(self):
        return self.label

    def get_title(self):
        return _(self.view_label).capitalize()

    def get_title_link(self):
        """Return title attribute for links to this view."""
        return self.title

    def get_title_html(self):
        """Return text for HTML title tag."""
        return self.title

    def get_title_heading(self):
        """Return text for page heading."""
        return self.title

    def get_template_name_suffix(self):
        return self.urlname

    def get_template_name_suffixes(self):
        return [self.template_name_suffix]

    def get_template_names(self):
        """Give a chance to default_template_name."""
        template_names = []
        template_name = getattr(self, 'template_name', None)
        if template_name:
            template_names.append(template_name)

        if getattr(self, 'model', None):
            template_names.append('%s/%s%s.html' % (
                self.model._meta.app_label,
                self.model._meta.model_name,
                '_' + self.urlname,
            ))

        default_template_name = getattr(self, 'default_template_name', None)
        if default_template_name:
            template_names.append(default_template_name)
        return template_names



from crudlfap.settings import *  # noqa

BASE_DIR = Path(__file__).resolve().parent.parent

INSTALLED_APPS += [  # noqa
    # CRUDLFA+ extras
    'django_registration',
    'crudlfap_registration',

    # CRUDLFA+ examples
    'crudlfap_example.artist',
    'crudlfap_example.song',
    'crudlfap_example.blog',
]

ROOT_URLCONF = os.path.dirname(__file__).split('/')[-1] + '.urls'

DATABASES = {
    'default': {
        'ENGINE': os.getenv('DB_ENGINE', 'django.db.backends.sqlite3'),
        'NAME': os.getenv('DB_NAME', os.path.join(
            os.path.dirname(__file__),
            '..',
            'db.sqlite3',
        )),
        'HOST': os.getenv('DB_HOST'),
        'USER': os.getenv('DB_USER'),
        'PASSWORD': os.getenv('DB_PASS'),
    }
}

install_optional(OPTIONAL_APPS, INSTALLED_APPS)  # noqa
install_optional(OPTIONAL_MIDDLEWARE, MIDDLEWARE)  # noqa

AUTHENTICATION_BACKENDS += [  # noqa
    'crudlfap_example.blog.crudlfap.AuthBackend',
]
